(*
 *
 * Copyright (c) 2011-2012,
 *  Alex Groce          <agroce@gmail.com>
 *  Chaoqiang Zhang     <zhangch@eecs.oregonstate.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
 
open Pretty
open Cil
open Cfg
open Usedef
open Printf
open String

module E=Errormsg

let sidbase = ref (-1)
let sidnextbase = ref(-1)

let cfginfofile = "cfginfofile"
let blkcountfile = "blockcount"

let typToStr t =
  (Pretty.sprint 1024 (Pretty.dprintf "%a" d_type t))

let expToStr e =
  (Pretty.sprint 4096 (Pretty.dprintf "%a" d_exp e))    
    
let stripPtr t =
  match t with
  | TPtr(t1,_) -> t1 
  | _ -> t                       

    
let rec changeList list =
  match list with 
    [] -> []
  | (str,t)::tail -> ((str,t,[])::(changeList tail))

let mkTypelist argtype =
  match argtype with
    [] -> None
  | _ -> (Some (changeList argtype))


let idCount = ref (-1)
let getIdCount () = ((idCount := !idCount + 1); !idCount)
let resetIdCount () = ( idCount := (-1); ) 


(* create a statement to call function fname 
with types argType and arguments args *)
let mkInstrFunCall fname argtype args =
  let fdec = emptyFunction fname
  in fdec.svar.vtype <- TFun(intType, (mkTypelist (("iid",intType)::argtype)), true, []);
  Call((None),(Lval(Var(fdec.svar),NoOffset)),((integer(getIdCount()))::args),locUnknown)


let prependInstrListToBody il b =
  List.iter (fun i -> b.bstmts <- (mkStmtOneInstr i) :: b.bstmts) 
    (List.rev il)

let d_cfgnodelabel (s : stmt) =
  begin
    match s.skind with
      | If (e, _, _, _)  -> "if " (*sprint ~width:999 (dprintf "if %a" d_exp e)*)
      | Loop _ -> "loop "
      | Break _ -> "break "
      | Continue _ -> "continue "
      | Goto _ -> "goto "
      | Instr _ -> "instr "
      | Switch _ -> "switch "
      | Block _ -> "block "
      | Return _ -> "return "
      | TryExcept _ -> "try-except "
      | TryFinally _ -> "try-finally "
  end
    
(* in a CFG, given a statement, 
    begin and end node -> 2
    begin node -> -1
    end node -> 1
    internal node -> 0
*)
let statStatement (s:stmt) = 
    ( match s.preds with
    | [] -> (match s.succs with 
                | [] -> 2
                | _  -> -1 )
    | _  -> (match s.succs with 
                | [] -> 1
                | _  -> 0 )
    )            

let hasLoopFollow(s:stmt) = 
    ( match s.succs with 
        | h::t -> ( let flag = statStatement(h) in (match h.skind with 
                        | Loop(_,_,_,_) -> [h.sid; statStatement(h)]
                        | _ -> [] ) )
        | _ -> []    
    )
    
    
let writeStr2File fname content = 
    (let chan = open_out_gen [Open_creat;Open_append] 33188 fname in
        begin
            ignore(fprintf chan "%s" content);
            close_out chan;
        end    )

let rec forallStmts (todo) (fd : fundec) = 
  begin
    fasBlock todo fd.sbody;
  end

and fasBlock (todo) (b : block) =
  List.iter (fasStmt todo) b.bstmts

and fasStmt (todo) (s : stmt) =
  begin
    (todo s);
    (*writeStr2File "log.txt" (d_cfgnodelabel s);*)
    match s.skind with
      | Block b -> ((*writeStr2File "log.txt" "a block\n";*) fasBlock todo b)
      | If (_, tb, fb, _) -> 
      ( (*writeStr2File "log.txt" (sprintf "length = %d;" (List.length tb.bstmts) );*) fasBlock todo tb; fasBlock todo fb)
      | Switch (_, b, _, _) -> fasBlock todo b
      | Loop (b, _, _, _) -> fasBlock todo b
      | (Return _ | Break _ | Continue _ | Goto _ | Instr _) -> ()
      | TryExcept _ | TryFinally _ -> E.s (E.unimp "try/except/finally")
  end
;;


(* print out cfg *)
let d_cfgnodename (s : stmt) =
  ( if !sidbase + s.sid > !sidnextbase  then sidnextbase := !sidbase + s.sid;
    sprintf "%d" (!sidbase+s.sid)
  )
let d_cfgedge (src) (dest) =
  sprintf "%s->%s;" (d_cfgnodename src) (d_cfgnodename dest)

let d_cfgnode (s : stmt) =
  sprintf "%s;" (d_cfgnodename s)

let d_cfgnode1tmp (s) (l) =
    String.concat "" (List.map (fun x -> (sprintf "%s" (d_cfgedge s x)) )l)
  
let d_cfgnode1 (s: stmt) =
    d_cfgnode1tmp s s.succs
        
(*let d_var() (s:stmt)*)
let firstnode = ref false
let funcparameters: VS.t ref = ref VS.empty


    
    
let d_node_du (s:stmt) = 
  begin  
    let varUsed, varDefs = Usedef.computeUseDefStmtKind ~acc_used:VS.empty ~acc_defs:VS.empty s.skind in 
    if( (!firstnode) == false ) 
    then 
        begin
            sprintf "%s(%s,|,%s);"
              (d_cfgnodename s)
              (VS.fold (fun x y -> let dd = sprintf "%s," x.vname in String.concat "" [dd;y]) varUsed "")
              (VS.fold (fun x y -> let dd = sprintf "%s," x.vname in String.concat "" [dd;y]) varDefs "")
        end
    else
        begin
        firstnode := false;
        let fulldefs = VS.union varDefs !funcparameters in
        sprintf "%s(%s,|,%s);"
          (d_cfgnodename s)
          (VS.fold (fun x y -> let dd = sprintf "%s," x.vname in String.concat "" [dd;y]) varUsed "")
          (VS.fold (fun x y -> let dd = sprintf "%s," x.vname in String.concat "" [dd;y]) fulldefs "")
        end
  end
  
(**********************************************************************)
(* entry points *)

(* print control flow graph (in dot form) for fundec to channel *)

let printCfgChannel (chan : out_channel) (fd : fundec) =
  let pnode (s:stmt) = fprintf chan "%s" (d_cfgnode s) in
  let pedge (s:stmt) = fprintf chan "%s" (d_cfgnode1 s) in
  let pdu (s:stmt) = fprintf chan "%s" (d_node_du s) in
    begin 
      ignore(fprintf chan "nodes: ");    
      forallStmts pnode fd;
      ignore(fprintf chan "\nedges: ");
      forallStmts pedge fd;  
      ignore(fprintf chan "\nduinfo: ");
      forallStmts pdu fd;
      ignore(fprintf chan "\n");
    end

(****************************************)
let printDuInfo(chan: out_channel) (fd: fundec) = 
  let duStmt (file: out_channel) ( s: stmt) =     
      (  let varUsed, varDefs = Usedef.computeUseDefStmtKind s.skind in 
           begin
             ignore(fprintf file "%s" "\nused variables\n");             
             VS.iter (fun arg-> ignore(fprintf file "%s;" arg.vname; )) varUsed;
             ignore(fprintf file "%s" "\ndefined variables\n");
             VS.iter (fun arg-> ignore(fprintf file "%s;" arg.vname; )) varDefs;
             ignore(fprintf file "\n------------------------\n");
           end
       )
  in List.iter (duStmt chan) fd.sbody.bstmts
(****************************************)
let curFile =  ref ""
class allbbCovVisitor = object (self) inherit nopCilVisitor

  (* the current function we're in, so when we visit a call node
   * we know who is the caller *)
  val mutable curFunc: string = ""

  method vfunc (f ) = 
    sidbase := !sidnextbase+1;
    ( Cil.prepareCFG(f);
      Cil.computeCFGInfo f false;  
      
      curFunc <- f.svar.vname;
      (*
      writeStr2File "du.txt" ("du information\n");
      let duchan = open_out_gen [Open_creat;Open_append] 33188 "du.txt" in
        begin
          printDuInfo duchan f;
          close_out duchan;
        end;
      *)
           (*
      firstnode:=true;
      funcparameters := VS.empty;
      List.iter (fun arg -> funcparameters := VS.add arg !funcparameters) f.sformals;
      
      writeStr2File cfginfofile ("\nfunname: " ^ curFunc ^ "\n");    
      (* export nodes:a;b;c \n edges: a->b; b->c; *)
      let chan = open_out_gen [Open_creat;Open_append] 33188 cfginfofile in
        begin
          printCfgChannel chan f;
          close_out chan;
        end;*)
      ); 
      
    DoChildren   
  
  method vstmt(s:stmt) =  
     ( match s.skind with
        | Instr (b) ->
        (if (s.sid >= 0) then
        (let l = (mkInstrFunCall "visit_basicblock"
            [  ("fname", charPtrType); ("sid", intType); ("stage", intType)]
            [ mkString(curFunc); integer(!sidbase+s.sid); integer( statStatement(s) ) ]
              )
        in (let smt = mkStmt(Instr(l::b) ) in (smt.labels <- s.labels; ChangeTo smt) )   ) 
        else (let smt = mkStmt(Instr(b)) in ( smt.labels<-s.labels; ChangeTo smt))  );
        
       | Block( b )     
            -> (  
                  (* work around for loop node instrument. It may not works for code other than sglib.*)
                  let extra = hasLoopFollow(s) in 
                    begin
                        match extra with
                            | [loopid;loopfollower] -> (prependInstrListToBody
                                                        [ (mkInstrFunCall "visit_basicblock"
                                                        [ ("fname", charPtrType); ("sid", intType); ("stage", intType)]
                                                        [ mkString(curFunc); integer(!sidbase+loopid); integer(loopfollower) ]
                                                          )] b )
                            | _ -> ()
                    end    ;
                    (prependInstrListToBody
                        [ (mkInstrFunCall "visit_basicblock"
                        [ ("fname", charPtrType); ("sid", intType); ("stage", intType)]
                        [ mkString(curFunc); integer(!sidbase+s.sid); integer(statStatement(s)) ]
                          )] b );
                   );  
                   DoChildren
       | If (_,_,_,_)
       | Return(_,_)
       | Goto(_,_)
            ->       
                (if ( 1>0)  then (self#queueInstr[(mkInstrFunCall "visit_basicblock"
                    [ ("fname", charPtrType); ("sid", intType); ("stage", intType)]
                    [ mkString(curFunc); integer(!sidbase+s.sid); integer( statStatement(s) ) ]
                    )]) else () 
                ); 
                DoChildren
        
        | _ -> DoChildren          
      )
    
end

class normalizeIfStmtVisitor = object (self) inherit nopCilVisitor
  method vstmt(s) =
    match s.skind with
      | If (e, b1, b2, _) ->
          if (b1.bstmts == []) then b1.bstmts <- [mkEmptyStmt ()] ;
          if (b2.bstmts == []) then b2.bstmts <- [mkEmptyStmt ()] ;
          DoChildren
      | _ -> DoChildren
end

let feature : featureDescr = 
  { fd_name = "allbbCov";
    fd_enabled = ref false;
    fd_description = "generation of code to produce all basic blocks coverage";
    fd_extraopt = [];
    fd_doit = 
    (function (f: file) ->
     
     let lwVisitor = new normalizeIfStmtVisitor in
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f;
      
     let lwVisitor = new allbbCovVisitor in
     ( 
      curFile := f.fileName;
      writeStr2File cfginfofile ("\nfilename: " ^ !curFile);      
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f;
      writeStr2File blkcountfile ( string_of_int (!idCount+1));
      )
      );
    fd_post_check = true;
  } 

