(*
 *
 * Copyright (c) 2011-2012,
 *  Alex Groce          <agroce@gmail.com>
 *  Chaoqiang Zhang     <zhangch@eecs.oregonstate.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
 

open Pretty
open Cil
open Printf

module H = Hashtbl

let writeStr2File fname content = 
  (let chan = open_out_gen [Open_creat;Open_wronly; Open_append; Open_text] 33188 fname in
  begin
    ignore(fprintf chan "%s" content);
    close_out chan;
  end  )
     
let rec changeList list =
  match list with 
    [] -> []
  | (str,t)::tail -> ((str,t,[])::(changeList tail))
   
let mkTypelist argtype =
  match argtype with
    [] -> None
  | _ -> (Some (changeList argtype))
         
(* create a statement to call function fname 
   with types argType and arguments args *)
let mkInstrFunCall fname argtype args =
  let fdec = emptyFunction fname in
  let cilid = -1
  in fdec.svar.vtype <- TFun(intType, (mkTypelist (("iid",intType)::argtype)), true, []);
  Call((None),(Lval(Var(fdec.svar),NoOffset)),((integer(cilid))::args),locUnknown)
     
(* collect all functions defined in this project*)
let functable : (string, int) H.t = H.create 10000
let funcid = ref (-1)
let getnewfuncid () = 
    ( funcid := (!funcid)+1; !funcid)
  
class functionsVisitor = object (self) inherit nopCilVisitor
    method vfunc (f ) = ( H.add functable (f.svar.vname) (getnewfuncid()); SkipChildren)
end

(* instrument call/return trace *)
class funcallretVisitor = object(self) inherit nopCilVisitor
  val mutable currfuncid: int = -1
      
  method vfunc (f) = 
    currfuncid <- H.find functable f.svar.vname;
    DoChildren
    
  (* visit an instruction; we're only interested in calls *)
  method vinst (i:instr) : instr list visitAction = begin
    match i with      
        | Call(_,f,_,_) -> 
            begin
                match f with
                 | Lval(Var(vi),NoOffset) -> 
                    begin
                        try 
                            let calleeidx = H.find functable vi.vname in  
                              let callinstr = (mkInstrFunCall "record_call_ret"
                                [ ("type", intType); ("src", intType); ("dest", intType)]
                                [ integer(0); integer(currfuncid); integer(calleeidx)])
                              and retinstr = (mkInstrFunCall "record_call_ret"
                                [ ("type", intType); ("src", intType); ("dest", intType)]
                                [ integer(1); integer(calleeidx);integer(currfuncid)])
                            in
                              ChangeTo ([callinstr;i;retinstr])
                        with Not_found -> DoChildren
                    end
                 | _ -> DoChildren
            end
        | _ -> DoChildren
  end
end

let feature : featureDescr = 
  { fd_name = "callRetCov";
    fd_enabled = ref false;
    fd_description = "generation of code to produce dynamic call return relationship";
    fd_extraopt = [];
    fd_doit = 
    (function (f: file) ->
      (* collect all functions name *)
      let funcVisitor = new functionsVisitor in
          visitCilFileSameGlobals (funcVisitor :> cilVisitor) f;
      (* output function list and their ids*)
      let s = H.fold (fun a b c -> (string_of_int b)^","^a ^ "\n" ^ c) functable "" in 
        writeStr2File "func_id_map" s;
      (* instrumentation for function call *)
      let callretVisitor = new funcallretVisitor in
          visitCilFileSameGlobals (callretVisitor :> cilVisitor) f;
    );       
      
    fd_post_check = true;
  } 
    
