(*
 *
 * Copyright (c) 2011-2012,
 *  Alex Groce          <agroce@gmail.com>
 *  Chaoqiang Zhang     <zhangch@eecs.oregonstate.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
 
open Pretty
open Cil
open Printf

module H = Hashtbl

module StringSet = Set.Make( 
  struct
    let compare = Pervasives.compare
    type t = string
  end )


let blockcount = ref (-1)

let globtable : (varinfo, int) H.t = H.create 100

let preddone = ref (StringSet.empty)

let globcount = ref (-1)

let predlist = ref ([])

let passdone = ref false

let inc_globcount () =
  (globcount := (!globcount)+1; !globcount)

let read_blockcount () =
  try
    let f = open_in "blockcount" in
    blockcount := (int_of_string (input_line f));
    (close_in f)
  with x ->
    blockcount := (-1)

let inc_blockcount () =
  (blockcount := (!blockcount)+1; !blockcount)

let write_blockcount () =
  try
    let f = open_out "blockcount" in
    output_string f (string_of_int (!blockcount+1));
    (*print_endline (string_of_int !blockcount);*)
    (close_out f)
  with x ->
    (prerr_string "Failed to write blockcount\n") ;;


let typToStr t =
  (Pretty.sprint 1024 (Pretty.dprintf "%a" d_type t))
    
let expToStr e =
  (Pretty.sprint 4096 (Pretty.dprintf "%a" d_exp e))  
    
let stripPtr t =
  match t with
  | TPtr(t1,_) -> t1 
  | _ -> t               
	
	
let rec changeList list =
  match list with 
    [] -> []
  | (str,t)::tail -> ((str,t,[])::(changeList tail))
	
let mkTypelist argtype =
  match argtype with
    [] -> None
  | _ -> (Some (changeList argtype))
	
	
let idCount = ref (-1)
let getIdCount () = ((idCount := !idCount + 1); !idCount)
let resetIdCount () = ( idCount := (-1); ) 
    
let curFile =  ref ""
    
(* create a statement to call function fname 
   with types argType and arguments args *)
let mkInstrFunCall fname argtype args =
  let fdec = emptyFunction fname
  in fdec.svar.vtype <- TFun(intType, (mkTypelist (("iid",intType)::argtype)), true, []);
  Call((None),(Lval(Var(fdec.svar),NoOffset)),((integer(getIdCount()))::args),locUnknown)
    
    
let prependInstrListToBody il b =
  List.iter (fun i -> b.bstmts <- (mkStmtOneInstr i) :: b.bstmts) 
    (List.rev il)
    
(*
let prependStmtListToBody il b =
  List.iter (fun i -> b.bstmts <- i :: b.bstmts) 
    (List.rev il)
*)

let mkInstrsStmts sl =
  List.flatten (List.map (fun i -> 
    (match i.skind with 
      Instr il -> il
    | _ -> [])) sl)

let stmtToStr e =
  (Pretty.sprint 4096 (Pretty.dprintf "%a" d_stmt e))  

let prependIl il1 il2 =
  List.flatten (List.map (fun i -> il1 @ [i]) il2)

let prependStmtListToBody sl b =
  let isl = mkInstrsStmts sl in
  b.bstmts <- (List.flatten (List.map (fun i -> 
    (match i.skind with
      Instr il -> [mkStmt (Instr (prependIl isl il))]
    | _ -> sl @ [i])) b.bstmts))
    
let writeStr2File fname content = 
  (let chan = open_out_gen [Open_creat;Open_append] 33188 fname in
  begin
    ignore(fprintf chan "%s" content);
    close_out chan;
  end  )
    
type check = PtrRef of exp | ArrRef of lval

let rec collectPointersLhost lhost l =
  match lhost with
  | Var v -> l
  | Mem e -> collectPointers e (PtrRef(e)::l)

and collectPointersOffset offset l =
  match offset with
  | NoOffset -> l
  | Index(e,offset1) -> collectPointers e  (collectPointersOffset offset1 l)
  | Field(f,offset1) -> collectPointersOffset offset1 l
	
and collectPointersLval lval l =
    match lval with
    | (lhost,Index(e2,offset1)) ->
	collectPointersLhost lhost (ArrRef(lval)::(collectPointersOffset offset1 l))
    | (lhost,offset) ->
	collectPointersLhost lhost (collectPointersOffset offset l)
    
and collectPointers e l =
  match e with
  | Const(c) -> l
  | Lval(lval) -> collectPointersLval lval l
  | SizeOf(t) -> l
  | SizeOfE(e1) -> l
  | SizeOfStr(s) -> l
  | AlignOf(t) -> l
  | AlignOfE(e1) -> l
  | UnOp(u,e1,t) -> collectPointers e1 l
  | BinOp(u,e1,e2,t) -> collectPointers e1 (collectPointers e2 l)
  | CastE(t,e1) -> collectPointers e1 l
  | AddrOf(lval) -> collectPointersLval lval l
  | StartOf(lval) -> collectPointersLval lval l
	
let rec makeChecks l =
  match l with 
  | [(PtrRef h)] -> BinOp(Ne, h, integer(0), Cil.intType)
  | (PtrRef h)::t -> BinOp(LAnd,BinOp(Ne, h, integer(0), Cil.intType),makeChecks t, Cil.intType)
  | (ArrRef h)::t -> makeChecks t
  | [(ArrRef h)] -> one
  | _ -> one
	
	
let rec inScopeLhost lhost cf =
  match lhost with
  | Var v -> inScopeVarinfo v cf
  | Mem e -> inScope e cf
	
and inScopeVarinfo v (f, cf) =
  ((v.vglob && ((H.find globtable v) <= !globcount)) || (cf == f))
    
and inScopeOffset offset cf =
  match offset with
  | NoOffset -> true
  | Index(e,offset1) -> (inScope e cf) && (inScopeOffset offset1 cf)
  | Field(f,offset1) -> inScopeOffset offset1 cf
	
and inScopeLval (lhost,offset) cf =
  (inScopeLhost lhost cf) && (inScopeOffset offset cf)
    
and inScope e cf =
  match e with
  | Const(c) -> true
  | Lval(lval) -> inScopeLval lval cf
  | SizeOf(t) -> true
  | SizeOfE(e1) -> inScope e1 cf
  | SizeOfStr(s) -> true
  | AlignOf(t) -> true
  | AlignOfE(e1) -> inScope e1 cf
  | UnOp(u,e1,t) -> inScope e1 cf
  | BinOp(u,e1,e2,t) -> (inScope e1 cf) && (inScope e2 cf)
  | CastE(t,e1) -> inScope e1 cf
  | AddrOf(lval) -> inScopeLval lval cf 
  | StartOf(lval) -> inScopeLval lval cf
	
let rec mapPredCalls (bc, cf, l) =
  match l with
  | (f, p)::t ->
      if (inScope p (f, cf) && (not (StringSet.mem (expToStr p) (!preddone)))) then
	inScopeCalls (bc, cf, p, t)
      else mapPredCalls (bc, cf, t)
  | [] -> []
	
and inScopeCalls (bc, cf, p, t) =
  let h =
    (let cv = collectPointers p [] in
    (if cv == [] then
      (preddone := (StringSet.add (expToStr p) (!preddone));
       mkStmtOneInstr( mkInstrFunCall "add_pred_value"
			[("pretty", charPtrType); ("value", intType)]
			[mkString(expToStr p); p ]))
    else 
      let safe = (makeChecks cv) in
      let safes = mkStmtOneInstr (mkInstrFunCall "add_pred_value"
				    [("pretty", charPtrType); ("value", intType)]
				    [mkString(expToStr p); p ]) in
      let nsafes = mkStmtOneInstr (mkInstrFunCall "add_pred_value"
				     [("pretty", charPtrType); ("value", intType)]
				     [mkString(expToStr p); integer 0 ]) in
      let b1 = mkBlock [safes] in
      let b2 = mkBlock [nsafes] in
      (preddone := (StringSet.add (expToStr p) (!preddone));
       mkStmt (If (safe, b1, b2, locUnknown))))) in
  h:: mapPredCalls(bc, cf, t)
      
let checkOffset e =
  match e with 
  | (Lval(e,offset)) -> ()
  | _ -> ()
  
      
let mkPredCalls bc cf l =
  (preddone := StringSet.empty;
   mkStmtOneInstr (mkInstrFunCall "begin_pred_values"
		     [ ("srcname", charPtrType); ("fname", charPtrType); ("bid", intType); ("tore", intType)]
		     [ mkString(!curFile); mkString(cf); integer(bc); integer(0)]))::mapPredCalls(bc, cf, l)
										       
class blockCovVisitor = object (self) inherit nopCilVisitor
    
  (* the current function we're in, so when we visit a call node
     * we know who is the caller *)
  val mutable curFunc: string = ""
      
  method vfunc (f ) = 
    curFunc <- f.svar.vname;
    DoChildren 
      
  method vglob (g) =
    ((match g with
    | GVarDecl (v, l) ->  (H.add globtable v (inc_globcount()))
    | GVar (v, i, l) -> (H.add globtable v (inc_globcount()))
    | _ -> ());
     DoChildren)
      
  method vexpr(e) =
    ((match e with
    | (Lval(Mem e1, offset)) -> (predlist := (curFunc, BinOp(Ne, e1, integer 0, Cil.intType))::(!predlist);
				 checkOffset (e))
    | (Lval(e1,offset)) -> checkOffset (e)
    | _ -> ());
     DoChildren)
      
  method vstmt(s) =
    ((if (not (!passdone)) then
      (match s.skind with
      | If(e,b1,b2,loc) ->
          (predlist := (curFunc,e)::(!predlist))
      | _ -> ())
    else ()); 
     DoChildren)
      
  method vblock(b) = 
    if ((!passdone) || (hasAttribute "predCover" b.battrs) ) then
      (if List.length b.bstmts > 0 then begin
        (match b.bstmts with
        | s::t ->
            (match s.skind with
            | Instr [Call(None,f,_,_)] ->
                (match f with
                | Lval(Var(fdecsvar),NoOffset) -> 
		    if ((compare fdecsvar.vname "add_pred_value") == 0) then ()
		    else prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b
		| _ -> prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b)
            | _ -> prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b)
        | _ -> prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b) end
      else (); )
      (*ignore(writeStr2File coverage_stmts_all ((string_of_int(!blockcount))^","));*)
    else ();
    DoChildren
      
end
    
let feature : featureDescr = 
  { fd_name = "predCovPtrStmt";
    fd_enabled = ref false;
    fd_description = "generation of code to produce predicate coverage at statement level, plus pointers/arrays";
    fd_extraopt = [];
    fd_doit = 
    (function (f: file) ->
      let lwVisitor = new blockCovVisitor in
      ( (*read_blockcount(); single file instrument, if you want to instrumetn multiple files in one session, uncomment this*)
      curFile := f.fileName;
    (*writeStr2File coverage_stmts_all ("\nfilename: " ^ !curFile);*)
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f;
      passdone := true;
      globcount := ( -1 );
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f));
    fd_post_check = true;
  } 
    
