(*
 *
 * Copyright (c) 2011-2012,
 *  Alex Groce          <agroce@gmail.com>
 *  Chaoqiang Zhang     <zhangch@eecs.oregonstate.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
 
open Pretty
open Cil
open Printf

module H = Hashtbl

module StringSet = Set.Make( 
  struct
    let compare = Pervasives.compare
    type t = string
  end )

let checkspacefuncs = [
"yaffs_start_up";
"yaffs_mount";
(*"yaffs_mount2";*)
"yaffs_unmount";
(*"yaffs_unmount2";*)
(*"yaffs_remount";*)
"yaffs_open";
"yaffs_close";
"yaffs_mkdir";
"yaffs_rmdir";
"yaffs_lseek";
"yaffs_truncate";
"yaffs_ftruncate";
"yaffs_stat";
"yaffs_fstat";
"yaffs_lstat";
"yaffs_read";
"yaffs_write";
"yaffs_opendir";
"yaffs_readdir";
"yaffs_rewinddir";
"yaffs_closedir";
"yaffs_link";
"yaffs_symlink";
"yaffs_readlink";
"yaffs_unlink";
"yaffs_rename";
"yaffs_chmod";
"yaffs_fchmod";
"yaffs_fsync";
"yaffs_fdatasync";
"yaffs_access";
"yaffs_dup";
"yaffs_pread";
"yaffs_pwrite";
"yaffs_utime";
"yaffs_futime";
"yaffs_flush";
"yaffs_sync";
"yaffs_inodecount";
"yaffs_nhandles"
]

let writeStr2File fname content = 
  (let chan = open_out_gen [Open_creat;Open_wronly; Open_append; Open_text] 33188 fname in
  begin
    ignore(fprintf chan "%s" content);
    close_out chan;
  end  )
  
let blockcount = ref (-1)

let globtable : (varinfo, int) H.t = H.create 100

let preddone = ref (StringSet.empty)

let globcount = ref (-1)

let predlist = ref ([])

let passdone = ref false

let inc_globcount () =
  (globcount := (!globcount)+1; !globcount)

let read_blockcount () =
  try
    let f = open_in "blockcount" in
    blockcount := (int_of_string (input_line f));
    (close_in f)
  with x ->
    blockcount := (-1)

let inc_blockcount () =
  (blockcount := (!blockcount)+1; !blockcount)

let write_blockcount () =
  try
    let f = open_out "blockcount" in
    output_string f (string_of_int (!blockcount+1));
    (*print_endline (string_of_int !blockcount);*)
    (close_out f)
  with x ->
    (prerr_string "Failed to write blockcount\n") ;;


let typToStr t =
  (Pretty.sprint 1024 (Pretty.dprintf "%a" d_type t))
    
let expToStr e =
  (Pretty.sprint 4096 (Pretty.dprintf "%a" d_exp e))  
    
let stripPtr t =
  match t with
  | TPtr(t1,_) -> t1 
  | _ -> t               
   
   
let rec changeList list =
  match list with 
    [] -> []
  | (str,t)::tail -> ((str,t,[])::(changeList tail))
   
let mkTypelist argtype =
  match argtype with
    [] -> None
  | _ -> (Some (changeList argtype))
   
   
let idCount = ref (-1)
let getIdCount () = ((idCount := !idCount + 1); !idCount)
let resetIdCount () = ( idCount := (-1); ) 

(* instrument point id*)
let idInstrPntCnt = ref(-1)
let getInstrPntIdCount () = ((idInstrPntCnt := !idInstrPntCnt + 1); !idInstrPntCnt)
let resetInstrPntIdCount () = ( idInstrPntCnt := (-1); ) 
    
let curFile =  ref ""
let idflag = ref (1)    
(* create a statement to call function fname 
   with types argType and arguments args *)
let mkInstrFunCallWithRet eret fname argtype args =
  let fdec = emptyFunction fname in
  let cilid = (if ((compare fname "add_pred_value") == 0) then
            getIdCount()
           else if ((compare fname "begin_pred_values") == 0) then
            getInstrPntIdCount()
           else
            -1) 
  in fdec.svar.vtype <- TFun(intType, (mkTypelist (argtype)), true, []);
  Call( (Some (Var eret,NoOffset)),(Lval(Var(fdec.svar),NoOffset)),(args),locUnknown)
  
let mkInstrFunCall fname argtype args =
  let fdec = emptyFunction fname in
  let cilid = (if ((compare fname "add_pred_value") == 0) then
            !idflag*getIdCount()
           else if ((compare fname "begin_pred_values") == 0) then
            !idflag*getInstrPntIdCount()
           else
            -1) 
  in fdec.svar.vtype <- TFun(intType, (mkTypelist (("iid",intType)::argtype)), true, []);
  Call((None),(Lval(Var(fdec.svar),NoOffset)),((integer(cilid))::args),locUnknown)
    
    
let prependInstrListToBody il b =
  List.iter (fun i -> b.bstmts <- (mkStmtOneInstr i) :: b.bstmts) 
    (List.rev il)
    
let prependStmtListToBody il b =
  List.iter (fun i -> b.bstmts <- i :: b.bstmts) 
    (List.rev il)
    
(* collect all functions defined in this project*)
let functable : (string, int) H.t = H.create 10000
let funcid = ref (-1)
let getnewfuncid () = 
    ( funcid := (!funcid)+1; !funcid)
  
class functionsVisitor = object (self) inherit nopCilVisitor
    method vfunc (f ) = ( H.add functable (f.svar.vname) (getnewfuncid()); SkipChildren)
end
    
type check = PtrRef of exp | ArrRef of lval

let rec collectPointersLhost lhost l =
  match lhost with
  | Var v -> l
  | Mem e -> collectPointers e (PtrRef(e)::l)

and collectPointersOffset offset l =
  match offset with
  | NoOffset -> l
  | Index(e,offset1) -> collectPointers e  (collectPointersOffset offset1 l)
  | Field(f,offset1) -> collectPointersOffset offset1 l
   
and collectPointersLval lval l =
    match lval with
    | (lhost,Index(e2,offset1)) ->
   collectPointersLhost lhost (ArrRef(lval)::(collectPointersOffset offset1 l))
    | (lhost,offset) ->
   collectPointersLhost lhost (collectPointersOffset offset l)
    
and collectPointers e l =
  match e with
  | Const(c) -> l
  | Lval(lval) -> collectPointersLval lval l
  | SizeOf(t) -> l
  | SizeOfE(e1) -> l
  | SizeOfStr(s) -> l
  | AlignOf(t) -> l
  | AlignOfE(e1) -> l
  | UnOp(u,e1,t) -> collectPointers e1 l
  | BinOp(u,e1,e2,t) -> collectPointers e1 (collectPointers e2 l)
  | CastE(t,e1) -> collectPointers e1 l
  | AddrOf(lval) -> collectPointersLval lval l
  | StartOf(lval) -> collectPointersLval lval l
   
let rec makeChecks l =
  match l with 
  | [(PtrRef h)] -> BinOp(Ne, h, integer(0), Cil.intType)
  | (PtrRef h)::t -> BinOp(LAnd,BinOp(Ne, h, integer(0), Cil.intType),makeChecks t, Cil.intType)
  | [(ArrRef h)] -> one
(*
      (match h with 
      | (lhost,Index(e2,offset)) ->
     (Printf.printf "ARRAY = %s\n" (expToStr (Lval(lhost,NoOffset)));
      try 
       (let s = lenOfArray (Some (Lval(lhost,NoOffset))) in
       (Printf.printf "2ARR e = %s length = %d\n" (expToStr (Lval(lhost,NoOffset))) s;
        BinOp(Lt, e2, integer s, Cil.intType)))
     with LenOfArray -> (Printf.printf "2BOMB\n"; zero))
      |    _ -> one)
*)
  | (ArrRef h)::t -> makeChecks t
  | [] -> one
   
   
let rec inScopeLhost lhost cf =
  match lhost with
  | Var v -> inScopeVarinfo v cf
  | Mem e -> false (* inScope e cf *)
   
and inScopeVarinfo v (f, cf) =
  ((v.vglob && ((H.find globtable v) <= !globcount)) || (cf == f))
    
and inScopeOffset offset cf =
  match offset with
  | NoOffset -> true
  | Index(e,offset1) -> false (* (inScope e cf) && (inScopeOffset offset1 cf) *)
  | Field(f,offset1) -> false (* inScopeOffset offset1 cf *)
   
and inScopeLval (lhost,offset) cf =
  (inScopeLhost lhost cf) && (inScopeOffset offset cf)
    
and inScope e cf =
  match e with
  | Const(c) -> true
  | Lval(lval) -> inScopeLval lval cf
  | SizeOf(t) -> true
  | SizeOfE(e1) -> inScope e1 cf
  | SizeOfStr(s) -> true
  | AlignOf(t) -> true
  | AlignOfE(e1) -> inScope e1 cf
  | UnOp(u,e1,t) -> inScope e1 cf
  | BinOp(u,e1,e2,t) -> (inScope e1 cf) && (inScope e2 cf)
  | CastE(t,e1) -> inScope e1 cf
  | AddrOf(lval) -> inScopeLval lval cf 
  | StartOf(lval) -> inScopeLval lval cf
   
let rec mapPredCalls (bc, cf, l) =
  match l with
  | (f, p)::t ->
      if (inScope p (f, cf) && (not (StringSet.mem (expToStr p) (!preddone)))) then
   inScopeCalls (bc, cf, p, t)
      else mapPredCalls (bc, cf, t)
  | [] -> []
   
and inScopeCalls (bc, cf, p, t) =
  let h =
    (let cv = collectPointers p [] in
    (if cv == [] then
      (preddone := (StringSet.add (expToStr p) (!preddone));
       mkStmtOneInstr( mkInstrFunCall "add_pred_value"
            [("pretty", charPtrType); ("value", intType)]
            [mkString(expToStr p); p ]))
    else 
      let safe = (makeChecks cv) in
      let safes = mkStmtOneInstr (mkInstrFunCall "add_pred_value"
                    [("pretty", charPtrType); ("value", intType)]
                    [mkString(expToStr p); p ]) in
      let nsafes = mkStmtOneInstr (mkInstrFunCall "add_pred_value"
                     [("pretty", charPtrType); ("value", intType)]
                     [mkString(expToStr p); integer 0 ]) in
      let b1 = mkBlock [safes] in
      let b2 = mkBlock [nsafes] in
      (preddone := (StringSet.add (expToStr p) (!preddone));
       mkStmt (If (safe, b1, b2, locUnknown))))) in
       (*writeStr2File "coverage.predicates.all" (string_of_int(bc)^"::"^(expToStr p)^"\n");*)
  h:: mapPredCalls(bc, cf, t)
      
let rec checkOffset cf e =
()
(*
  match e with 
  | (Lval(lhost,Index(e2,offset))) ->
      (try
   (let s = lenOfArray (Some e) in
   (Printf.printf "ARR e = %s length = %d\n" (expToStr e) s;
    predlist := (cf, BinOp(Lt, e2, integer s, Cil.intType))::(!predlist)))
      with LenOfArray -> (Printf.printf ("BOMB\n")))
  | _ -> ()
*)
  
class viFinderClass br = object(self)
      inherit nopCilVisitor
      
      method vvrbl vi' = 
      (br := true;
      SkipChildren)
      
end
      
let expContainsVi e  =
     let br = ref false in
     let vis = new viFinderClass  br in
     ignore(visitCilExpr vis e);
     !br
     
(*-----------------*)

let globs : (varinfo, int) H.t = H.create 10000
     
class globvCollector = object(self) inherit nopCilVisitor
    method vglob (g) =
    ((match g with
    | GVarDecl (v, l) ->  (H.add globs v 0)
    | GVar (v, i, l) -> (H.add globs v 0)
    | _ -> ());
     DoChildren)
end

class varFinderClass br = object(self)
      inherit nopCilVisitor
      
      method vvrbl vi' = 
      ( br := vi'::!br;
      DoChildren)
      
end

let expAllAreGVi e  =
     let vlist = ref [] in
        let vis = new varFinderClass vlist in
            visitCilExpr vis e;
     let allg = ref true in     
     let tmpf i = (try 
                    H.find globs i; ()
                with
                    Not_found -> allg := false; )
        in
    List.iter (fun i -> tmpf i) !vlist;
    !allg

(*-----------------*)     
let mkPredCalls bc cf l =
  (preddone := StringSet.empty;
   mkStmtOneInstr (mkInstrFunCall "begin_pred_values"
           [ ("srcname", charPtrType); ("fname", charPtrType); ("bid", intType); ("tore", intType)]
           [ mkString(!curFile); mkString(cf); integer(bc); integer(0)]))::mapPredCalls(bc, cf, l)

let currFuncDesc =    ref []    

class predFuncEntryCovVisitor = object (self) inherit nopCilVisitor
    
  (* the current function we're in, so when we visit a call node
     * we know who is the caller *)
  val mutable curFunc: string = ""
  val mutable binstrumented: bool = false 
  val mutable currfuncid: int = -1
  method vfunc (f ) = 
    currfuncid <- H.find functable f.svar.vname;
    currFuncDesc := [];
    currFuncDesc := f::!currFuncDesc;
    binstrumented <- false;
    curFunc <- f.svar.vname;
    DoChildren 
      
  method vglob (g) =
    ((match g with
    | GVarDecl (v, l) ->  (H.add globtable v (inc_globcount()))
    | GVar (v, i, l) -> (H.add globtable v (inc_globcount()))
    | _ -> ());
     DoChildren)
      
(*  method vexpr e =
    (Printf.printf "EXP: %s  " (expToStr (e));
     (try
       (let s = lenOfArray (Some (e))
       in Printf.printf "Size: %d\n" s)
     with LenOfArray -> Printf.printf "No length\n");
     DoChildren)
*)

  method vexpr(e) =
    ((match e with
    | (Lval(Mem e1, offset)) -> if expAllAreGVi e1 then
                                    (predlist := (curFunc, BinOp(Ne, e1, integer 0, Cil.intType))::(!predlist);
                                    checkOffset curFunc e)
                                else ()
    | (Lval(e1,offset)) -> checkOffset curFunc e
    | _ -> ());
     DoChildren)
      
  method vstmt(s) =
    ((if (not (!passdone)) then
      (match s.skind with
      | If(e,b1,b2,loc) ->
            if expAllAreGVi e then
                (predlist := (curFunc,e)::(!predlist))
            else
                ()
      | _ -> ())
    else ()); 
     DoChildren)
      
  method vblock(b) = 
    if ( ( not binstrumented ) &&(List.exists (fun it -> it = curFunc ) checkspacefuncs)&& ( (!passdone) || (hasAttribute "predCover" b.battrs) ) ) then
      (binstrumented <- true;
      if List.length b.bstmts > 0 then 
        begin
            (match b.bstmts with
            | s::t ->
                (match s.skind with
                | Instr [Call(None,f,_,_)] ->
                    (match f with
                    | Lval(Var(fdecsvar),NoOffset) -> 
              if ((compare fdecsvar.vname "add_pred_value") == 0) then ()
              else prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b
            | _ -> prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b)
                | _ -> prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b)
            | _ -> prependStmtListToBody (mkPredCalls (inc_blockcount()) curFunc !predlist) b);
             (if List.exists (fun it -> it = curFunc ) checkspacefuncs 
                then
                 let t = makeTempVar (List.hd !currFuncDesc) intType in
                 prependInstrListToBody 
                 [ mkInstrFunCall "cil_yaffs_freespace" [ ("func", intType);("path", charPtrType)] [integer(currfuncid); mkString "/yaffs2"] ] b
                else ()
                );            
        end
      else (); )
      (*ignore(writeStr2File coverage_stmts_all ((string_of_int(!blockcount))^","));*)
    else ();
    DoChildren
      
end

class predIfEntryCovVisitor = object (self) inherit nopCilVisitor    

   method vstmt(s:stmt) = 
    (match s.skind with
      | If(e,_,_,_) -> 
        if expContainsVi e then
            ( self#queueInstr[mkInstrFunCall "add_pred_value"
              [("pretty", charPtrType); ("value", intType)]
              [mkString(expToStr e); e ]];
            DoChildren)
        else
            DoChildren
        
      | _ -> DoChildren
    );    
    DoChildren    
end
    
let feature : featureDescr = 
  { fd_name = "predFuncEntryCov";
    fd_enabled = ref false;
    fd_description = "generation of code to produce predicate coverage on function entries";
    fd_extraopt = [];
    fd_doit = 
    (function (f: file) ->
      idflag := -1;
      let funcVisitor = new functionsVisitor in
        visitCilFileSameGlobals (funcVisitor :> cilVisitor) f;
      let s = H.fold (fun a b c -> (string_of_int b)^","^a ^ "\n" ^ c) functable "" in 
        writeStr2File "func_id_map" s;
        
      let gvVisitor = new globvCollector in
              visitCilFileSameGlobals (gvVisitor :> cilVisitor) f;
      
      let lwVisitor = new predFuncEntryCovVisitor in
      ( (*read_blockcount(); single file instrument, if you want to instrumetn multiple files in one session, uncomment this*)
      curFile := f.fileName;

      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f;
      
      predlist := List.filter (fun e -> expContainsVi (snd e) ) !predlist;
         
     let diffpreds = (List.map (fun i -> (fst i)^"::"^(expToStr (snd i))) !predlist) in
       let s = List.fold_left (fun a b -> a ^ b ^ "\n") "" diffpreds in
        writeStr2File "diffpreds" s;

     passdone := true;
      globcount := ( -1 );
     
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f);  
   idflag := 1;
     let lwVisitor = new predIfEntryCovVisitor in
      ( (*read_blockcount(); single file instrument, if you want to instrumetn multiple files in one session, uncomment this*)
      curFile := f.fileName;
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f); 
      
     let f = open_out "pred_func_entry_num" in
      output_string f (string_of_int(!idCount+1));
      (close_out f);
   );       
      
    fd_post_check = true;
  } 
    
