(*
 *
 * Copyright (c) 2011-2012,
 *  Alex Groce          <agroce@gmail.com>
 *  Chaoqiang Zhang     <zhangch@eecs.oregonstate.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
 
open Pretty
open Cil
open Printf

module H = Hashtbl

module StringSet = Set.Make( 
  struct
    let compare = Pervasives.compare
    type t = string
  end )

let writeStr2File fname content = 
  (let chan = open_out_gen [Open_creat;Open_append] 33188 fname in
  begin
    ignore(fprintf chan "%s" content);
    close_out chan;
  end  )

let gcurrFunc = ref ""
let blockid = ref (-1)
let stmttable : (stmt ref, stmt ref) H.t = H.create 10000
let globtable : (varinfo, int) H.t = H.create 100

let preddone = ref (StringSet.empty)

let globid = ref (-1)

let predlist = ref ([])

let passdone = ref false

let getNextGlobId () =
  (globid := (!globid)+1; !globid)

let getNextBlockId () =
  (blockid := (!blockid)+1; !blockid)
    
let expToStr e =
  (Pretty.sprint 4096 (Pretty.dprintf "%a" d_exp e))  
    
let predid = ref (-1)
let getNextPredId () = ((predid := !predid + 1); !predid)
    
let curFile =  ref ""
    
let callCilAddPredValue args = 
    let fdec = emptyFunction "__cil_add_pred_value" in
    fdec.svar.vtype <- TFun(voidType, Some [("id",intType,[]);("pretty",charPtrType,[]);("value",intType,[])], false, []);
    Call(None, Lval(Var(fdec.svar),NoOffset),args, locUnknown)

let callCilBeginPredValues args = 
    let fdec = emptyFunction "__cil_begin_pred_values" in
    fdec.svar.vtype <- TFun(voidType, Some [("id",intType,[]);("srcname", charPtrType,[]);("fname",charPtrType,[])], false, []);
    Call(None, Lval(Var(fdec.svar),NoOffset),args, locUnknown)
    
let mkInstrsStmts sl =
  List.flatten (List.map (fun i -> 
    (match i.skind with 
      Instr il -> il
    | _ -> [])) sl)

let stmtToStr e =
  (Pretty.sprint 4096 (Pretty.dprintf "%a" d_stmt e))  

let prependIl il1 il2 =
  List.flatten (List.map (fun i -> il1 @ [i]) il2)
                
type check = PtrRef of exp | ArrRef of lval

let rec collectPointersLhost lhost l =
  match lhost with
  | Var v -> l
  | Mem e -> collectPointers e (PtrRef(e)::l)

and collectPointersOffset offset l =
  match offset with
  | NoOffset -> l
  | Index(e,offset1) -> collectPointers e  (collectPointersOffset offset1 l)
  | Field(f,offset1) -> collectPointersOffset offset1 l
    
and collectPointersLval lval l =
    match lval with
    | (lhost,Index(e2,offset1)) ->
    collectPointersLhost lhost (ArrRef(lval)::(collectPointersOffset offset1 l))
    | (lhost,offset) ->
    collectPointersLhost lhost (collectPointersOffset offset l)
    
and collectPointers e l =
  match e with
  | Const(c) -> l
  | Lval(lval) -> collectPointersLval lval l
  | SizeOf(t) -> l
  | SizeOfE(e1) -> l
  | SizeOfStr(s) -> l
  | AlignOf(t) -> l
  | AlignOfE(e1) -> l
  | UnOp(u,e1,t) -> collectPointers e1 l
  | BinOp(u,e1,e2,t) -> collectPointers e1 (collectPointers e2 l)
  | CastE(t,e1) -> collectPointers e1 l
  | AddrOf(lval) -> collectPointersLval lval l
  | StartOf(lval) -> collectPointersLval lval l
    
let rec makeChecks l =
  match l with 
  | [(PtrRef h)] -> BinOp(Ne, h, integer(0), Cil.intType)
  | (PtrRef h)::t -> BinOp(LAnd,BinOp(Ne, h, integer(0), Cil.intType),makeChecks t, Cil.intType)
  | [(ArrRef h)] -> one
(*
      (match h with 
      | (lhost,Index(e2,offset)) ->
      (Printf.printf "ARRAY = %s\n" (expToStr (Lval(lhost,NoOffset)));
       try 
        (let s = lenOfArray (Some (Lval(lhost,NoOffset))) in
        (Printf.printf "2ARR e = %s length = %d\n" (expToStr (Lval(lhost,NoOffset))) s;
         BinOp(Lt, e2, integer s, Cil.intType)))
      with LenOfArray -> (Printf.printf "2BOMB\n"; zero))
      |     _ -> one)
*)
  | (ArrRef h)::t -> makeChecks t
  | [] -> one
    
    
let rec inScopeLhost lhost cf =
  match lhost with
  | Var v -> inScopeVarinfo v cf
  | Mem e -> false (* inScope e cf *)
    
and inScopeVarinfo v (f, cf) =
  ((v.vglob && ((H.find globtable v) <= !globid)) || (cf == f))
    
and inScopeOffset offset cf =
  match offset with
  | NoOffset -> true
  | Index(e,offset1) -> false (* (inScope e cf) && (inScopeOffset offset1 cf) *)
  | Field(f,offset1) -> false (* inScopeOffset offset1 cf *)
    
and inScopeLval (lhost,offset) cf =
  (inScopeLhost lhost cf) && (inScopeOffset offset cf)
    
and inScope e cf =
  match e with
  | Const(c) -> true
  | Lval(lval) -> inScopeLval lval cf
  | SizeOf(t) -> true
  | SizeOfE(e1) -> inScope e1 cf
  | SizeOfStr(s) -> true
  | AlignOf(t) -> true
  | AlignOfE(e1) -> inScope e1 cf
  | UnOp(u,e1,t) -> inScope e1 cf
  | BinOp(u,e1,e2,t) -> (inScope e1 cf) && (inScope e2 cf)
  | CastE(t,e1) -> inScope e1 cf
  | AddrOf(lval) -> inScopeLval lval cf 
  | StartOf(lval) -> inScopeLval lval cf
    
let rec mapPredCalls (bc, cf, l) =
  match l with
  | (f, p)::t ->
      if (inScope p (f, cf) && (not (StringSet.mem (expToStr p) (!preddone)))) then
    inScopeCalls (bc, cf, p, t)
      else mapPredCalls (bc, cf, t)
  | [] -> []
    
and inScopeCalls (bc, cf, p, t) =
  let h =
    (let cv = collectPointers p [] in
    (if cv == [] then
      (preddone := (StringSet.add (expToStr p) (!preddone));
       mkStmtOneInstr( callCilAddPredValue [integer (getNextPredId()); mkString(expToStr p); p ]))
    else 
      let safe = (makeChecks cv) in
      let safes = mkStmtOneInstr (callCilAddPredValue [integer (getNextPredId()); mkString(expToStr p); p ]) in
      let nsafes = mkStmtOneInstr (callCilAddPredValue [integer (getNextPredId()); mkString(expToStr p); p ]) in
      let b1 = mkBlock [safes] in
      let b2 = mkBlock [nsafes] in
      (preddone := (StringSet.add (expToStr p) (!preddone));
       mkStmt (If (safe, b1, b2, locUnknown))))) in
         h:: mapPredCalls(bc, cf, t)
      
let rec checkOffset cf e =
()
(*
  match e with 
  | (Lval(lhost,Index(e2,offset))) ->
      (try
    (let s = lenOfArray (Some e) in
    (Printf.printf "ARR e = %s length = %d\n" (expToStr e) s;
     predlist := (cf, BinOp(Lt, e2, integer s, Cil.intType))::(!predlist)))
      with LenOfArray -> (Printf.printf ("BOMB\n")))
  | _ -> ()
*)
  
class viFinderClass br = object(self)
      inherit nopCilVisitor
      
      method vvrbl vi' = 
      (br := true;
      SkipChildren)
      
end
      
let expContainsVi e  =
     let br = ref false in
     let vis = new viFinderClass  br in
     ignore(visitCilExpr vis e);
     !br
     
let mkPredCalls bc cf l =
  (preddone := StringSet.empty;
   mkStmtOneInstr (callCilBeginPredValues [integer(bc); mkString(!curFile); mkString(cf) ]))::mapPredCalls(bc, cf, l)

let addmap bs news = 
  match bs with
  | [] -> ()
  | x::xs -> match x.labels with
              | [] -> ()
              | _ -> H.add stmttable (ref x) (ref (List.hd news)); (List.hd news).labels <- x.labels
              
              
let prependStmtListToBody sl b =
  (*let isl = mkInstrsStmts sl in*)

  b.bstmts <- (List.flatten (List.map (fun i -> 
    (match i.skind with
      Instr il -> let kk = mkStmt (Instr (List.flatten (List.map (fun j -> (mkInstrsStmts (mkPredCalls (getNextBlockId()) !gcurrFunc !predlist)) @ [j] ) il))) 
        in kk.labels <- i.labels; [kk]
    | _ -> (mkPredCalls (getNextBlockId()) !gcurrFunc !predlist) @ [i])) b.bstmts))

class fixGotoVisitor = object (self) inherit nopCilVisitor 
  method vstmt(s) = 
    (match s.skind with
      | Goto (t,p) -> (try 
                         let v = H.find stmttable t in
                         ChangeTo (mkStmt (Goto (v, p)) )
                      with _ -> SkipChildren )
      | _ -> DoChildren
    )
end
class fixGoto1Visitor = object (self) inherit nopCilVisitor 
  method vstmt(s) = 
    match s.labels with
      | [] -> DoChildren
      | _  -> ( try 
                let v = H.find stmttable (ref s) in
                  s.labels <- [];
                  DoChildren
                with _ -> DoChildren  
              ) 
end
let callBeginInstrument args= 
    let fdec = emptyFunction "begin_trace_session" in
    fdec.svar.vtype <- TFun(voidType, Some [("typename", charPtrType,[]);("bid", intType,[])], false, []);
    Call(None, Lval(Var(fdec.svar),NoOffset),args, locUnknown)

let prependInstrToBlk il b =
  (b.bstmts <- let news = (mkStmtOneInstr il) in news:: b.bstmts)

class addCovStarterVisitor = object (self) inherit nopCilVisitor   

  val mutable curFunc: string = ""
  val mutable blockID: int = 0        
  method vfunc (f ) = 
    (curFunc <- f.svar.vname;
    DoChildren )
      
  method vblock(b) = 
      ( if String.compare (curFunc) "main" == 0 then  begin
         if blockID==0 then
           begin prependInstrToBlk
            ( callBeginInstrument [(mkString "predStmt"); integer 0] ) b ; 
            blockID <- blockID+1;
           end;
         end;         
     );
     DoChildren
      
end

class predStmtCovVisitor = object (self) inherit nopCilVisitor   

  val mutable curFunc: string = ""
  
  method vfunc (f ) = 
    curFunc <- f.svar.vname;
    gcurrFunc := f.svar.vname;
    DoChildren 
      
  method vglob (g) =
    ((match g with
    | GVarDecl (v, l) ->  (H.add globtable v (getNextGlobId()))
    | GVar (v, i, l) -> (H.add globtable v (getNextGlobId()))
    | _ -> ());
     DoChildren)
      
(*  method vexpr e =
    (Printf.printf "EXP: %s  " (expToStr (e));
     (try
       (let s = lenOfArray (Some (e))
       in Printf.printf "Size: %d\n" s)
     with LenOfArray -> Printf.printf "No length\n");
     DoChildren)
*)

  method vexpr(e) =
    ((match e with
    | (Lval(Mem e1, offset)) -> (predlist := (curFunc, BinOp(Ne, e1, integer 0, Cil.intType))::(!predlist);
                 checkOffset curFunc e)
    | (Lval(e1,offset)) -> checkOffset curFunc e
    | _ -> ());
     DoChildren)
      
  method vstmt(s) =
    ((if (not (!passdone)) then
      (match s.skind with
      | If(e,b1,b2,loc) ->
          (predlist := (curFunc,e)::(!predlist))
      | _ -> ())
    else ()); 
     DoChildren)
      
  method vblock(b) =  
    if ((!passdone) || (hasAttribute "predCover" b.battrs) ) then
      (if List.length b.bstmts > 0 then begin
        (match b.bstmts with
        | s::t ->
            (match s.skind with
            | Instr [Call(None,f,_,_)] ->
                (match f with
                | Lval(Var(fdecsvar),NoOffset) -> 
            if ((compare fdecsvar.vname "__cil_add_pred_value") == 0) then ()
            else prependStmtListToBody 0 b
        | _ -> prependStmtListToBody 0 b)
            | _ -> prependStmtListToBody 0 b)
        | _ -> prependStmtListToBody 0 b) end
      else (); 
        )
    else ();
    DoChildren
      
end
    
let feature : featureDescr = 
  { fd_name = "predStmtCov";
    fd_enabled = ref false;
    fd_description = "generation of code to produce predicate coverage at statement levelx";
    fd_extraopt = [];
    fd_doit = 
    (function (f: file) ->
      let lwVisitor = new predStmtCovVisitor in
      (       

      curFile := f.fileName;
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f;
      
      predlist := List.filter (fun e -> expContainsVi (snd e) ) !predlist;
            
      passdone := true;
      globid := ( -1 );
      visitCilFileSameGlobals (lwVisitor :> cilVisitor) f); 

      let starter1Visitor = new addCovStarterVisitor in
  			visitCilFileSameGlobals (starter1Visitor :> cilVisitor) f;

      let gotoVisitor = new fixGotoVisitor in
       visitCilFileSameGlobals (gotoVisitor :> cilVisitor) f;
      let goto1Visitor = new fixGoto1Visitor in
       visitCilFileSameGlobals (goto1Visitor :> cilVisitor) f;



      let f = open_out "stmtpredcount" in
          output_string f (string_of_int(!predid+1));
        (close_out f);
    );
    fd_post_check = true;
  } 
