/*
 * covhooks.cpp: coverage instrumentation hooks.
 *
 * Copyright (C) 2011-2013 Oregon State University.
 *
 * Created by Chaoqiang Zhang <zhangch@eecs.oregonstate.edu>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <vector>
#include <list>
#include <string>
#include <stack>
#include <map>
#include <algorithm>

/////////////////////////////////////////
// predefine consts for all coverage

int cblocks = -1;
int cbranches = -1;
int cpredicates = -1;
int cstmtpredicates = -1;


enum eCOVTYPE {
	BLOCK = 1, BRANCH, PREDICATE, STMTPREDICATE, IMP, AIMP
};

eCOVTYPE g_covtype;

typedef std::vector<int> ti_path;
typedef std::vector<ti_path> ti_paths;

std::stack<ti_path, std::list<std::vector<int> > > currpaths;
std::stack<std::string, std::list<std::string> > currfuncs;
typedef std::map<std::string, ti_paths> func_paths_map;
typedef std::map<std::string, ti_paths>::const_iterator const_func_paths_map_iterator;
typedef std::map<std::string, ti_paths>::iterator func_paths_map_iterator;
typedef std::vector<ti_path>::const_iterator const_paths_iterator;
func_paths_map m_func_paths_map;

FILE* g_covfile = NULL;
unsigned char* g_covdata = NULL;

int readuint(const char* filename) {
	int num = -1;
	FILE* fp = fopen(filename, "r");
	if (fp) {
		fscanf(fp, "%d", &num);
		fclose(fp);
	} else {
		//assert(0);
	}
	return num;
}

int writeuint(const char* filename, int d) {
	FILE* fp = fopen(filename, "w");
	if (fp) {
		fprintf(fp, "%d", d);
		fclose(fp);
	} else {
		assert(0);
	}
	return 0;
}

extern "C" void end_trace_session();

extern "C" void begin_trace_session(char* type, int addnew) {
  atexit(end_trace_session);
	if (strcmp(type, "block") == 0) {
    printf("block");
		g_covtype = BLOCK;
		cblocks = readuint("blockcount");
		if (!g_covdata)
			g_covdata = (unsigned char*) malloc(cblocks);
		memset(g_covdata, 0, cblocks);
	} else if (strcmp(type, "branch") == 0) {
		g_covtype = BRANCH;
		cbranches = readuint("branchcount");
		if (!g_covdata)
			g_covdata = (unsigned char*) malloc(cbranches);
		memset(g_covdata, 0, cbranches);
	} else if (strcmp(type, "path") == 0) {
		g_covtype = IMP;
		m_func_paths_map.clear();
	} else if (strcmp(type, "acyclicpath") == 0) {
		g_covtype = AIMP;
		m_func_paths_map.clear();
	} else if (strcmp(type, "pred") == 0) {
		g_covtype = PREDICATE;
		cpredicates = readuint("blockpredcount");
		if (!g_covdata)
			g_covdata = (unsigned char*) malloc(cpredicates);
		memset(g_covdata, 0, cpredicates);
	} else if (strcmp(type, "predStmt") == 0) {
		g_covtype = STMTPREDICATE;
		cstmtpredicates = readuint("stmtpredcount");
		if (!g_covdata)
			g_covdata = (unsigned char*) malloc(cstmtpredicates);
		memset(g_covdata, 0, cstmtpredicates);
	}
}

void writepathfile(const func_paths_map& data, const char* pathfile) {
  int totalpaths=0;
  FILE* currfile = fopen(pathfile, "a+");
  const_func_paths_map_iterator itr = data.begin();
  for(; itr != data.end(); itr++ ) {
      const char* name = itr->first.c_str();
     fprintf(currfile, "[%s]\n", name);
      const ti_paths& paths = itr->second;
      ti_paths::const_iterator pathsitr = paths.begin();
      for( ; pathsitr != paths.end(); pathsitr++ ) {
          totalpaths++;
          const ti_path& path = *pathsitr;
          ti_path::const_iterator pathitr = path.begin();
          for( ; pathitr != path.end(); pathitr++ ) {
              int tmp = *pathitr;
              fprintf(currfile, "%d,", tmp);
          }
          fprintf(currfile, "\n");
      }        
  }
  fclose(currfile);
}

extern "C" void add_func_path_entry(const std::string& func_name,
		ti_path& newpath) {
	func_paths_map_iterator itr = m_func_paths_map.find(func_name);
	if (itr != m_func_paths_map.end()) {
		ti_paths& paths4func = itr->second;
		bool bfound = false;
		for (int i = 0; i < paths4func.size(); i++) {
			if (paths4func[i].size() == newpath.size()) {
				if ((paths4func[i] == newpath)) {
					bfound = true;
					break;
				}
			}
		}
		if (!bfound) {
			paths4func.push_back(newpath);
		}
	} else {
		ti_paths tmppaths;
		tmppaths.clear();
		tmppaths.push_back(newpath);
		m_func_paths_map.insert(std::make_pair(func_name, tmppaths));
	}
}

extern "C" void end_trace_session() {
	if (g_covtype == BLOCK) {
		g_covfile = fopen("block.cov", "a+");
		int i = 0;
		for (i = 0; i < cblocks; i++) {
			fprintf(g_covfile, "%d,", g_covdata[i]);
		}
		fprintf(g_covfile, "\n");
		fclose(g_covfile);
	} else if (g_covtype == BRANCH) {
		g_covfile = fopen("branch.cov", "a+");

		int i = 0;
		for (i = 0; i < cbranches; i++) {
			fprintf(g_covfile, "%d,", g_covdata[i]);
		}
		fprintf(g_covfile, "\n");

		fclose(g_covfile);
	} else if (g_covtype == IMP) {
		while( currpaths.size() > 0) {
			std::vector<int>& currpath = currpaths.top();
			std::string& f = currfuncs.top();
		   add_func_path_entry(f, currpath);
		   currpaths.pop();
		   currfuncs.pop();
	   }		
		int x = readuint("pathcount");
		writeuint("pathcount", x+1);
		char fname[128];      		
		sprintf(fname, "path%06d.cov", x+1);       
      writepathfile(m_func_paths_map, fname);
	} else if (g_covtype == AIMP) {
		while( currpaths.size() > 0) {
			std::vector<int>& currpath = currpaths.top();
			std::string& f = currfuncs.top();
		   add_func_path_entry(f, currpath);
		   currpaths.pop();
		   currfuncs.pop();
	   }
		int x = readuint("acyclicpathcount");
		writeuint("acyclicpathcount", x+1);
		char fname[128];      		
		sprintf(fname, "acyclicpath%06d.cov", x+1);
		writepathfile(m_func_paths_map, fname);
	} else if (g_covtype == PREDICATE) {
		// write an integer
		g_covfile = fopen("pred.cov", "a+");
		int i = 0;

		for (i = 0; i < cpredicates; i++) {
			int low = 0;
			int high = 0;
			if (g_covdata[i] == 1) {
				low = 1;
			} else if (g_covdata[i] == 2) {
				high = 1;
			} else if (g_covdata[i] == 3) {
				low = 1;
				high = 1;
			}
			fprintf(g_covfile, "%d,%d,", low, high);
		}

		fprintf(g_covfile, "\n");
		fclose(g_covfile);
	} else if (g_covtype == STMTPREDICATE) {
		g_covfile = fopen("predStmt.cov", "a+");

		int i = 0;

		for (i = 0; i < cstmtpredicates; i++) {
			int low = 0;
			int high = 0;
			if (g_covdata[i] == 1) {
				low = 1;
			} else if (g_covdata[i] == 2) {
				high = 1;
			} else if (g_covdata[i] == 3) {
				low = 1;
				high = 1;
			}
			fprintf(g_covfile, "%d,%d,", low, high);
		}

		fprintf(g_covfile, "\n");
		fclose(g_covfile);
	}
}


extern "C" void __cil_hit_block(char* fname, char* mname, int bid) {
	if (g_covdata)
		g_covdata[bid] = 1;
}
extern "C" void __cil_hit_branch(char* fname, char* mname, int bid) {
	if (g_covdata)
		g_covdata[bid] = 1;
}

extern "C" void __cil_hit_cfgnode(char* fname, char* mname, int bid, int stat) {
	std::string currfunc = mname;
	/* start new function path */
	if (bid == 0 /*stat == -1 || stat == 2*/) {
		std::vector<int> newpath;
		newpath.clear();
		currpaths.push(newpath);
		currfunc = mname;
		currfuncs.push(currfunc);
	}
	/* intermediate nodes */
	if (stat == -1 || stat == 0 || stat == 1 || stat == 2) {
		assert( !currpaths.empty());
		std::vector<int>& currpath = currpaths.top();
		if (g_covtype == AIMP) {
			if (std::find(currpath.begin(), currpath.end(), bid)
					!= currpath.end()) {
				add_func_path_entry(currfunc, currpath);
				currpath.clear();
			}
		}
		currpath.push_back(bid);
	}
	/* function exit */
	if (stat == 1 || stat == 2) {
		assert( !currpaths.empty());
		std::vector<int>& currpath = currpaths.top();
		add_func_path_entry(currfunc, currpath);

		currpaths.pop();
		currfuncs.pop();
	}
}

extern "C" void __cil_begin_pred_values(int blkid, char* filename,
		char* funcname) {
}

extern "C" void __cil_add_pred_value(int id, char* predicate, int predvalue) {
	unsigned char d = predvalue != 0;
	if (g_covdata)
		g_covdata[id] |= (d ? 2 : 1 );
}

